==> fotoalbum-0.33.tar.bz2: fotoalbum-0.33/src/shared/readjpeg.cpp <==
/**
 * This file is largely based on FLTK's Fl_JPEG_Image.cxx, which says:
 * 
// Copyright 1997-2005 by Easy Software Products.
// Image support donated by Matthias Melcher, Copyright 2000.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
 *
 * Modified by Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 */

#include "readjpeg.h"
#include "../config.h"

#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>


// Some releases of the Cygwin JPEG libraries don't have a correctly
// updated header file for the INT32 data type; the following define
// from Shane Hill seems to be a usable workaround...

#if defined(WIN32) && defined(__CYGWIN__)
#  define XMD_H
#endif // WIN32 && __CYGWIN__


extern "C"
{
#ifdef USE_LIBJPEG
#  include <jpeglib.h>
#endif // USE_LIBJPEG
}


//
// Custom JPEG error handling structure...
//

#ifdef USE_LIBJPEG
struct fljw_jpeg_error_mgr {
  jpeg_error_mgr        pub_;           // Destination manager...
  jmp_buf               errhand_;       // Error handler
};
#endif // USE_LIBJPEG


//
// Error handler for JPEG files...
//

#ifdef USE_LIBJPEG
extern "C" {
  static void
  fljw_jpeg_error_handler(j_common_ptr dinfo) { // I - Decompressor info
    longjmp(((fljw_jpeg_error_mgr *)(dinfo->err))->errhand_, 1);
  }

  static void
  fljw_jpeg_output_handler(j_common_ptr) {      // I - Decompressor info (not used)
  }
}
#endif // USE_LIBJPEG


unsigned char* LoadScaledJPEG(const char *jpeg, int& nWMin, int& nHMin)
{
  // Clear data...
  unsigned char* array = 0;
  
#ifdef USE_LIBJPEG
  FILE                          *fp;    // File pointer
  jpeg_decompress_struct        dinfo;  // Decompressor info
  fljw_jpeg_error_mgr           jerr;   // Error handler info
  JSAMPROW                      row;    // Sample row pointer

  // the following variables are pointers allocating some private space that
  // is not reset by 'setjmp()'
  char* max_finish_decompress_err;      // count errors and give up afer a while
  char* max_destroy_decompress_err;     // to avoid recusion and deadlock


  // Open the image file...
  if ((fp = fopen(jpeg, "rb")) == NULL) 
    return(0);

  // Setup the decompressor info and read the header...
  dinfo.err                = jpeg_std_error((jpeg_error_mgr *)&jerr);
  jerr.pub_.error_exit     = fljw_jpeg_error_handler;
  jerr.pub_.output_message = fljw_jpeg_output_handler;

  // Setup error loop variables
  max_finish_decompress_err = (char*)malloc(1);   // allocate space on the frame for error counters
  max_destroy_decompress_err = (char*)malloc(1);  // otherwise, the variables are reset on the longjmp
  *max_finish_decompress_err=10;
  *max_destroy_decompress_err=10;

  if (setjmp(jerr.errhand_))
  {
    // JPEG error handling...
    // if any of the cleanup routines hits another error, we would end up 
    // in a loop. So instead, we decrement max_err for some upper cleanup limit.
    if ( ((*max_finish_decompress_err)-- > 0) && array)
      jpeg_finish_decompress(&dinfo);
    if ( (*max_destroy_decompress_err)-- > 0)
      jpeg_destroy_decompress(&dinfo);

    fclose(fp);

    if (array) 
      delete[] array;

    free(max_destroy_decompress_err);
    free(max_finish_decompress_err);

    return(0);
  }

  jpeg_create_decompress(&dinfo);
  jpeg_stdio_src(&dinfo, fp);
  jpeg_read_header(&dinfo, 1);
  
  // this is the real difference to the fltk code:
  {
    float fact = (float)dinfo.image_height / nHMin;
    if (fact < (float)dinfo.image_width / nWMin)
      fact = (float)dinfo.image_width / nWMin;
    if (fact >= 8)
      dinfo.scale_denom = 8;
    else if (fact >= 4)
      dinfo.scale_denom = 4;
    else if (fact >= 2)
      dinfo.scale_denom = 2;
    else
      dinfo.scale_denom = 1;
    dinfo.scale_num = 1;
    // printf("orig=%dx%d, fact=%d, lim=%dx%d\n", dinfo.image_width, dinfo.image_height, dinfo.scale_denom, nWMin, nHMin);
  }

  dinfo.quantize_colors      = (boolean)FALSE;
  dinfo.out_color_space      = JCS_RGB;
  dinfo.out_color_components = 3;
  dinfo.output_components    = 3;

  jpeg_calc_output_dimensions(&dinfo);
  
  int w = dinfo.output_width;
  int h = dinfo.output_height;
  int d = dinfo.output_components;

  nWMin = w;
  nHMin = h;
  
  array = new unsigned char[w * h * d];

  jpeg_start_decompress(&dinfo);

  while (dinfo.output_scanline < dinfo.output_height) {
    row = (JSAMPROW)(array +
                     dinfo.output_scanline * dinfo.output_width *
                     dinfo.output_components);
    jpeg_read_scanlines(&dinfo, &row, (JDIMENSION)1);
  }

  jpeg_finish_decompress(&dinfo);
  jpeg_destroy_decompress(&dinfo);

  free(max_destroy_decompress_err);
  free(max_finish_decompress_err);

  fclose(fp);
#else
    printf("USE_LIBJPEG fehlt\n");
#endif // USE_LIBJPEG
  
  return(array);
}

int WriteJPEG(const char* filename, unsigned char* data, int nW, int nH, int nD, int nQuality)
{
  struct jpeg_compress_struct cinfo;
  struct jpeg_error_mgr jerr;
  
  /* this is a pointer to one row of image data */
  JSAMPROW row_pointer[1];
  
  FILE* outfile = fopen(filename, "wb");  
  if (!outfile)
  {
    printf("Error opening output jpeg file %s\n!", filename);
    return -1;
  }
  
  cinfo.err = jpeg_std_error(&jerr);
  jpeg_create_compress(&cinfo);
  jpeg_stdio_dest(&cinfo, outfile);
  
  /* Setting the parameters of the output file here */
  cinfo.image_width = nW;
  cinfo.image_height = nH;
  cinfo.input_components = nD;
  cinfo.in_color_space = JCS_RGB;
  /* default compression parameters, we shouldn't be worried about these */
  jpeg_set_defaults(&cinfo);
  jpeg_set_quality(&cinfo, nQuality, 0);
  /* Now do the compression .. */
  jpeg_start_compress(&cinfo, TRUE);
  /* like reading a file, this time write one row at a time */
  while (cinfo.next_scanline < cinfo.image_height)
  {
    row_pointer[0] = &data[ cinfo.next_scanline * cinfo.image_width *  cinfo.input_components];
    jpeg_write_scanlines(&cinfo, row_pointer, 1);
  }
  /* similar to read file, clean up after we're done compressing */
  jpeg_finish_compress(&cinfo);
  jpeg_destroy_compress(&cinfo);
  fclose(outfile);
  
  return 0;
}

==> fotoalbum-0.201.tar.bz2: fotoalbum-0.201/src/shared/readjpeg.cpp <==
/**
 * This file is largely based on FLTK's Fl_JPEG_Image.cxx, which says:
 * 
// Copyright 1997-2005 by Easy Software Products.
// Image support donated by Matthias Melcher, Copyright 2000.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
 *
 * Modified by Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 */

#include "readjpeg.h"
#include "../config.h"

#include <stdio.h>
#include <iostream>
#include <stdlib.h>
#include <setjmp.h>
#include <errno.h>
#include <cstring>

#ifdef WIN32_FLTK13
# include <FL/fl_utf8.h>
#endif

// Some releases of the Cygwin JPEG libraries don't have a correctly
// updated header file for the INT32 data type; the following define
// from Shane Hill seems to be a usable workaround...

#if defined(WIN32) && defined(__CYGWIN__)
#  define XMD_H
#endif // WIN32 && __CYGWIN__


extern "C"
{
#ifdef USE_LIBJPEG
#  include <jpeglib.h>
#endif // USE_LIBJPEG
}


//
// Custom JPEG error handling structure...
//

#ifdef USE_LIBJPEG
struct fljw_jpeg_error_mgr {
  jpeg_error_mgr        pub_;           // Destination manager...
  jmp_buf               errhand_;       // Error handler
};
#endif // USE_LIBJPEG


//
// Error handler for JPEG files...
//

#ifdef USE_LIBJPEG
extern "C" {
  static void
  fljw_jpeg_error_handler(j_common_ptr dinfo) { // I - Decompressor info
    longjmp(((fljw_jpeg_error_mgr *)(dinfo->err))->errhand_, 1);
  }

  static void
  fljw_jpeg_output_handler(j_common_ptr) {      // I - Decompressor info (not used)
  }
}
#endif // USE_LIBJPEG


unsigned char* LoadScaledJPEG(const char *jpeg, int& nWMin, int& nHMin, int& nRatio)
{
  // Clear data...
  unsigned char* array = 0;
  
#ifdef USE_LIBJPEG
  FILE                          *fp;    // File pointer
  jpeg_decompress_struct        dinfo;  // Decompressor info
  fljw_jpeg_error_mgr           jerr;   // Error handler info
  JSAMPROW                      row;    // Sample row pointer

  // the following variables are pointers allocating some private space that
  // is not reset by 'setjmp()'
  char* max_finish_decompress_err;      // count errors and give up afer a while
  char* max_destroy_decompress_err;     // to avoid recusion and deadlock


  // Open the image file...
  #ifdef WIN32_FLTK13
    fp = fl_fopen(jpeg, "rb");
  #else
    fp = fopen(jpeg, "rb");
  #endif
  if (fp == NULL) 
  {
    int nErr = errno;
    std::cerr << "Error opening " << jpeg << ":" << strerror(nErr) << "\n";
    return(0);
  }

  // Setup the decompressor info and read the header...
  dinfo.err                = jpeg_std_error((jpeg_error_mgr *)&jerr);
  jerr.pub_.error_exit     = fljw_jpeg_error_handler;
  jerr.pub_.output_message = fljw_jpeg_output_handler;

  // Setup error loop variables
  max_finish_decompress_err = (char*)malloc(1);   // allocate space on the frame for error counters
  max_destroy_decompress_err = (char*)malloc(1);  // otherwise, the variables are reset on the longjmp
  *max_finish_decompress_err=10;
  *max_destroy_decompress_err=10;

  if (setjmp(jerr.errhand_))
  {
    // JPEG error handling...
    // if any of the cleanup routines hits another error, we would end up 
    // in a loop. So instead, we decrement max_err for some upper cleanup limit.
    if ( ((*max_finish_decompress_err)-- > 0) && array)
      jpeg_finish_decompress(&dinfo);
    if ( (*max_destroy_decompress_err)-- > 0)
      jpeg_destroy_decompress(&dinfo);

    fclose(fp);

    if (array) 
      delete[] array;

    free(max_destroy_decompress_err);
    free(max_finish_decompress_err);

    return(0);
  }

  jpeg_create_decompress(&dinfo);
  jpeg_stdio_src(&dinfo, fp);
  jpeg_read_header(&dinfo, 1);
  
  // this is the real difference to the fltk code:
  {
    float fact = (float)dinfo.image_height / nHMin;
    if (fact < (float)dinfo.image_width / nWMin)
      fact = (float)dinfo.image_width / nWMin;
    if (fact >= 8)
      dinfo.scale_denom = 8;
    else if (fact >= 4)
      dinfo.scale_denom = 4;
    else if (fact >= 2)
      dinfo.scale_denom = 2;
    else
      dinfo.scale_denom = 1;
    dinfo.scale_num = 1;
    nRatio = dinfo.scale_denom;
    // printf("orig=%dx%d, fact=%d, lim=%dx%d\n", dinfo.image_width, dinfo.image_height, dinfo.scale_denom, nWMin, nHMin);
  }

  dinfo.quantize_colors      = (boolean)FALSE;
  dinfo.out_color_space      = JCS_RGB;
  dinfo.out_color_components = 3;
  dinfo.output_components    = 3;

  jpeg_calc_output_dimensions(&dinfo);
  
  int w = dinfo.output_width;
  int h = dinfo.output_height;
  int d = dinfo.output_components;

  nWMin = w;
  nHMin = h;
  
  array = new unsigned char[w * h * d];

  jpeg_start_decompress(&dinfo);

  while (dinfo.output_scanline < dinfo.output_height) {
    row = (JSAMPROW)(array +
                     dinfo.output_scanline * dinfo.output_width *
                     dinfo.output_components);
    jpeg_read_scanlines(&dinfo, &row, (JDIMENSION)1);
  }

  jpeg_finish_decompress(&dinfo);
  jpeg_destroy_decompress(&dinfo);

  free(max_destroy_decompress_err);
  free(max_finish_decompress_err);

  fclose(fp);
#else
    printf("USE_LIBJPEG fehlt\n");
#endif // USE_LIBJPEG
  
  return(array);
}

int WriteJPEG(const char* filename, unsigned char* data, int nW, int nH, int nD, int nQuality)
{
  struct jpeg_compress_struct cinfo;
  struct jpeg_error_mgr jerr;
  
  /* this is a pointer to one row of image data */
  JSAMPROW row_pointer[1];
  
  FILE* outfile;
  #ifdef WIN32_FLTK13
    outfile = fl_fopen(filename, "wb");
  #else
    outfile = fopen(filename, "wb");
  #endif
  
  if (!outfile)
  {
    printf("Error opening output jpeg file %s\n!", filename);
    return -1;
  }
  
  cinfo.err = jpeg_std_error(&jerr);
  jpeg_create_compress(&cinfo);
  jpeg_stdio_dest(&cinfo, outfile);
  
  /* Setting the parameters of the output file here */
  cinfo.image_width = nW;
  cinfo.image_height = nH;
  cinfo.input_components = nD;
  cinfo.in_color_space = JCS_RGB;
  /* default compression parameters, we shouldn't be worried about these */
  jpeg_set_defaults(&cinfo);
  jpeg_set_quality(&cinfo, nQuality, 0);
  /* Now do the compression .. */
  jpeg_start_compress(&cinfo, TRUE);
  /* like reading a file, this time write one row at a time */
  while (cinfo.next_scanline < cinfo.image_height)
  {
    row_pointer[0] = &data[ cinfo.next_scanline * cinfo.image_width *  cinfo.input_components];
    jpeg_write_scanlines(&cinfo, row_pointer, 1);
  }
  /* similar to read file, clean up after we're done compressing */
  jpeg_finish_compress(&cinfo);
  jpeg_destroy_compress(&cinfo);
  fclose(outfile);
  
  return 0;
}
